#  Copyright 2023 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import cached_property
from typing import Dict, Tuple, TYPE_CHECKING

import numpy as np
from attrs import frozen
from cirq_ft import TComplexity

from qualtran import Bloq, Signature

if TYPE_CHECKING:
    import cirq

    from qualtran.cirq_interop import CirqQuregT


@frozen
class TGate(Bloq):
    """The T gate.

    This is the fourth root of the Pauli Z gate. Quantum programs composed solely
    of gates belonging to the Clifford group (like X, Z, Hadamard, CNOT, S) can be simulated
    efficiently by a classical computer and therefore offer no quantum advantage. The Clifford
    gates do not provide a universal quantum gateset. The addition of any non-Clifford gate
    makes the gateset universal. One of the most popular additions is the T gate, yielding
    the common Clifford+T gateset.

    Registers:
     - q: The qubit

    References:
        [Universal Quantum Computation with ideal Clifford gates and noisy ancillas](https://arxiv.org/abs/quant-ph/0403025).
            Bravyi and Kitaev. 2004.
        [Fast and efficient exact synthesis of single qubit unitaries generated by Clifford and T gates](https://arxiv.org/abs/1206.5236).
            Kliuchnikov et. al. 2012.
        [Universal Gate Set, Magic States, and costliness of the T gate](https://quantumcomputing.stackexchange.com/a/33358).
            Gidney. 2023.
    """

    @cached_property
    def signature(self) -> 'Signature':
        return Signature.build(q=1)

    def t_complexity(self) -> 'TComplexity':
        return TComplexity(t=1)

    def as_cirq_op(
        self, qubit_manager: 'cirq.QubitManager', q: 'CirqQuregT'
    ) -> Tuple['cirq.Operation', Dict[str, 'CirqQuregT']]:
        import cirq

        (q,) = q
        return cirq.T(q), {'q': np.array([q])}
